import os.path
import mimetypes
from multiprocessing.pool import ThreadPool as Pool
from multiprocessing import Queue

from tqdm import tqdm

import cv2

from pixelateme.face_detection import FaceDetection
from pixelateme.centerface import CenterFace
from pixelateme.blur import Blur


kill_preview = False
detection_buffer = None


def get_files(paths):
    result = []
    for path in paths:
        if os.path.isfile(path):
            result.append(path)
        else:
            for file in os.listdir(path):
                result.append(os.path.join(path, file))

    return result


def get_type(path):
    if not os.path.exists(path):
        return None

    mime = mimetypes.guess_type(path)
    if mime[0].startswith("video"):
        return "video"
    if mime[0].startswith("image"):
        return "image"
    return None


def init_pool(d_b):
    global detection_buffer
    detection_buffer = d_b


def scale_preview_image(image, width=720, height=480):
    border_v = 0
    border_h = 0
    if (height / width) >= (image.shape[0] / image.shape[1]):
        border_v = int((((height / width) * image.shape[1]) - image.shape[0]) / 2)
    else:
        border_h = int((((width / height) * image.shape[0]) - image.shape[1]) / 2)
    img = cv2.copyMakeBorder(image, border_v, border_v, border_h, border_h, cv2.BORDER_CONSTANT, 0)
    return cv2.resize(img, (width, height))


def get_blurred_frame(face_detection: FaceDetection, blur: Blur, frame, preview):
    global kill_preview
    global detection_buffer
    boxes = face_detection.get_boxes(frame)
    blurred = blur.blur_faces(image=frame, boxes=boxes)
    if preview and not kill_preview:
        if detection_buffer is None:
            cv2.imshow("Preview (press q to exit preview, blurring will continue)", scale_preview_image(blurred))
            a = cv2.waitKey(1)
            if a == ord('q'):
                kill_preview = True
                cv2.destroyAllWindows()
        else:
            detection_buffer.put(scale_preview_image(blurred))
    return blurred


def get_output_file_name(path, kwargs):
    basename = os.path.splitext(path)
    folder = os.path.join(os.path.abspath(kwargs.get("output")), basename[0] + kwargs.get("suffix") + basename[1])
    os.makedirs(os.path.dirname(folder), exist_ok=True)
    return folder


def process_image(path, face_detection: FaceDetection, blur: Blur, pbar, kwargs):
    frame = cv2.imread(path)
    blurred = get_blurred_frame(face_detection, blur, frame, kwargs.get("preview"))
    cv2.imwrite(get_output_file_name(path, kwargs), blurred)
    if pbar is not None:
        pbar.update(1)


def process_video(path, face_detection: FaceDetection, blur: Blur, pbar, kwargs):
    cap = cv2.VideoCapture(path)
    ret, img = cap.read()
    out = cv2.VideoWriter(get_output_file_name(path, kwargs), int(cap.get(cv2.CAP_PROP_FOURCC)),
                          cap.get(cv2.CAP_PROP_FPS),
                          (img.shape[1], img.shape[0]))
    while ret:
        blurred = get_blurred_frame(face_detection, blur, img, kwargs.get("preview"))
        out.write(blurred)

        if pbar is not None:
            pbar.update(1)

        ret, img = cap.read()

    cap.release()
    out.release()


def get_frame_amount(paths):
    frames = 0
    for path in paths:
        file_type = get_type(path)
        if file_type is None:
            continue
        if file_type == "video":
            cap = cv2.VideoCapture(path)
            frames += int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            cap.release()
        else:
            frames += 1
    return frames


def process_path(path, face_detection, blur, pbar, kwargs):
    file_type = get_type(path)

    if file_type is None:
        print(f"Warning: File {path} is not a video and not an image, skipping...")

    if file_type == "video":
        process_video(path, face_detection, blur, pbar, kwargs)
    else:
        process_image(path, face_detection, blur, pbar, kwargs)


def show():
    while True:
        frame = detection_buffer.get()
        if frame is not None:
            cv2.imshow("Preview (press q to exit preview, blurring will continue)", frame)
        else:
            break
        if cv2.waitKey(1) & 0xFF == ord('q'):
            cv2.destroyAllWindows()
            break
    return


def run(**kwargs):
    in_shape = None
    if kwargs.get("face_recognition_size") is not None:
        splitted = kwargs.get("face_recognition_size").split("x")
        in_shape = int(splitted[0]), int(splitted[1])
    centerface = CenterFace(backend=kwargs.get("backend"), in_shape=in_shape)
    all_except_images, only_this_images = [], []
    if kwargs.get("all_except_images") is not None:
        all_except_image_files = get_files([kwargs.get("all_except_images")])
        for path in all_except_image_files:
            all_except_images.append(cv2.imread(path))

    if kwargs.get("only_this_images") is not None:
        only_this_images_files = get_files([kwargs.get("only_this_images")])
        for path in only_this_images_files:
            all_except_images.append(cv2.imread(path))

    face_detection = FaceDetection(centerface, all_except_images_data=all_except_images,
                                   only_this_images_data=only_this_images, **kwargs)
    blur = Blur(**kwargs)

    paths = get_files(kwargs.get("path"))
    frame_amount = get_frame_amount(paths)
    pbar = None
    if frame_amount > 1:
        pbar = tqdm(total=frame_amount, desc="Processing frames...")

    if kwargs.get("multiprocessing"):
        buffer = Queue()
        pool = Pool(kwargs.get("parallel_processes") + 1, initializer=init_pool, initargs=(buffer,))
        if kwargs.get("preview"):
            show_future = pool.apply_async(show)
        futures = []
        for path in paths:
            f = pool.apply_async(process_path, args=(path, face_detection, blur, pbar, kwargs,))
            futures.append(f)

        for f in futures:
            f.get()
        if kwargs.get("preview"):
            global detection_buffer
            detection_buffer.put(None)
            show_future.get()
    else:
        for path in paths:
            process_path(path, face_detection, blur, pbar, kwargs)

    if kwargs.get("preview"):
        cv2.destroyAllWindows()



from deepface import DeepFace


class FaceDetection:
    def __init__(self, centerface, all_except_images_data=[], only_this_images_data=[], **kwargs):
        self.all_except_images = all_except_images_data
        self.only_this_images = only_this_images_data
        self.threshold = kwargs.get("threshold")
        self.deepface_similarity = kwargs.get("deepface_similarity")
        self.centerface = centerface

    def blur_face(self, face):
        for image in self.all_except_images:
            result = DeepFace.verify(face, image, enforce_detection=False)
            if result["distance"] <= self.deepface_similarity:
                return False

        blur = True
        for image in self.only_this_images:
            result = DeepFace.verify(face, image, enforce_detection=False)
            if result["distance"] <= self.deepface_similarity:
                return True
            blur = False

        return blur

    def filter_boxes(self, image, boxes):
        def condition(box):
            x1, y1, x2, y2 = box
            return self.blur_face(image[y1:y2, x1:x2])
        return list(filter(lambda x: condition(x), boxes))

    def get_boxes(self, frame):
        dets, _ = self.centerface(frame, threshold=self.threshold)
        boxes = list(map(lambda x: x[:4], dets))
        boxes = list(map(lambda x: list(map(lambda y: int(y), x)), boxes))

        return self.filter_boxes(frame, boxes)



import cv2
import numpy as np


class Blur:
    def __init__(self, mode="blur", ellipse=True, **kwargs):
        self.mode = mode
        self.ellipse = ellipse
        self.kwargs = kwargs

    @staticmethod
    def get_ellipse(box):
        x1, y1, x2, y2 = box
        center = x1 + (x2 - x1) // 2, y1 + (y2 - y1) // 2
        size = (x2 - x1) // 2, (y2 - y1) // 2
        return center, size

    @staticmethod
    def get_ellipse_masks(shape, boxes):
        mask = np.zeros(shape, dtype="uint8")
        for box in boxes:
            center, size = Blur.get_ellipse(box)
            cv2.ellipse(mask, center, size, 0, 0, 360, (255, 255, 255), -1)
        return mask

    @staticmethod
    def get_box_masks(shape, boxes):
        mask = np.zeros(shape, dtype="uint8")
        for box in boxes:
            x1, y1, x2, y2 = box
            cv2.rectangle(mask, (x1, y1), (x2, y2), (255, 255, 255), -1)
        return mask

    def get_mask(self, shape, boxes):
        mask = Blur.get_ellipse_masks(shape, boxes) if self.ellipse else Blur.get_box_masks(shape, boxes)
        if self.kwargs.get("soft_mask"):
            mask = cv2.GaussianBlur(mask, self.get_blur_kernel(self.kwargs.get("soft_mask_strength"), mask.shape), 0)
        return mask

    @staticmethod
    def get_blur_kernel(blur_strength, image_shape):
        kernel_size = int(blur_strength / 100 * image_shape[1]), int(blur_strength / 100 * image_shape[1])
        return tuple(map(lambda x: x if x % 2 == 1 else x + 1, kernel_size))

    def mix_images(self, source: np.array, blurred: np.array, mask: np.array):
        if not self.kwargs.get("soft_mask"):
            return np.where(mask > 0, blurred, source)
        mask = mask / 255
        result = blurred * mask + source * (1 - mask)
        return result.astype(np.uint8)

    def solid(self, image, boxes):
        mask = self.get_mask(image.shape, boxes)
        zeros = np.zeros(image.shape, dtype="uint8")
        return self.mix_images(image, zeros, mask)

    def blur(self, image, boxes):
        kernel_size = self.get_blur_kernel(self.kwargs.get("blur_strength"), image.shape)
        blurred = cv2.GaussianBlur(image, kernel_size, 0)
        mask = self.get_mask(image.shape, boxes)
        return self.mix_images(image, blurred, mask)

    def pixelate(self, image, boxes):
        mask = self.get_mask(image.shape, boxes)
        height, width = image.shape[:2]
        w, h = (self.kwargs.get("pixelate_size") / 100 * width, self.kwargs.get("pixelate_size") / 100 * height)
        temp = cv2.resize(image, (int(width / w), int(height / h)), interpolation=cv2.INTER_LINEAR)
        pixelate = cv2.resize(temp, (width, height), interpolation=cv2.INTER_NEAREST)

        return self.mix_images(image, pixelate, mask)

    def blur_faces(self, image, boxes):
        if self.mode == "color":
            return self.solid(image, boxes)
        if self.mode == "blur":
            return self.blur(image, boxes)
        if self.mode == "pixelate":
            return self.pixelate(image, boxes)

        raise Exception("Unsupported mode. Supported modes are color, blur and pixelate!")



from pixelateme.main import run
from gooey import Gooey, GooeyParser

@Gooey(progress_regex=r"^(\d+)/(\d+)$", progress_expr="x[0] / x[1] * 100")
def parse_args():
    args = GooeyParser(description="Pixelate all faces on a image or video")
    args.add_argument("path", nargs="*", help="File path(s) and folders containing images and/or videos to pixelate", widget="MultiFileChooser")
    out = args.add_argument_group("Output")
    out.add_argument("--suffix", default="", help="A suffix which is concatenated to processed files")
    out.add_argument("--output", "-o", default="./pixelated",
                      help="Output directory, where all pixelated files are saved at", widget="DirChooser")
    blur_group = args.add_argument_group("Anonymization", "Configure anonymization settings")
    blur_group.add_argument("--mode", "-m", default="pixelate", choices=["pixelate", "blur", "color"], help="Anonymization mode")
    blur_group.add_argument("--ellipse", action="store_true",
                      help="Use ellipses as form to pixelate face instead of a rectangle")
    blur_group.add_argument("--blur-strength", dest="blur_strength", type=float, default=3,
                      help="Strength of the blur effect. The higher the value, the stronger the blur. Only working when mode is blur")
    blur_group.add_argument("--pixelate-size", dest="pixelate_size", type=float, default=5,
                      help="Raster size of pixelated effect. The low the value, the more blocks are generated. Only working when mode is pixelate")
    blur_group.add_argument("--soft-mask", dest="soft_mask", action="store_true", help="Enables a soft transition between blurred and original image")
    blur_group.add_argument("--soft-mask-strength", dest="soft_mask_strength", default=7, type=float, help="Defines the feather strength of the mask edge")

    face_detection = args.add_argument_group("Face Detection")
    face_detection.add_argument("--threshold", "-t", default=0.5, type=float, help="Threshold for face recognition", widget="DecimalField")
    face_detection.add_argument("--backend", default="auto", choices=["auto", "onnxrt", "opencv"], help="Backend for face recognition")
    face_detection.add_argument("--face-recognition-size", dest="face_recognition_size", default=None,
                                help="Maximum size of image used for face recognition in the format WxH (e.g. 720x480). Larger images/videos will be downscaled. Has no affect on output resolution.")
    face_detection.add_argument("--maximum-face-recognition-size", dest="max_face_recognition_size", default=640,
                                type=int,
                                help="The maximum number of pixels of the longest side. Image/Video will be scaled down according longest side.")

    face_recognition = args.add_argument_group("Face Recognition")
    face_recognition.add_argument("--only-blur-this-faces", default=None, dest="only_this_images", help="Folder containing images with faces. Only faces matching with one in this folder are pixelated", widget="DirChooser")
    face_recognition.add_argument("--blur-except-this-faces", default=None, dest="all_except_images", help="Folder containing images with faces. Only faces, that don't match any face in the folder, are pixelated", widget="DirChooser")
    face_recognition.add_argument("--deepface-similarity", dest="deepface_similarity", type=float, default=0.4, help="Maximum cosinus similarity of two faces, that are considered as 'same face'. Only relevant with either only-blur-this-faces or blur-except-this-faces set.")
    args.add_argument("--preview", dest="preview", action="store_true", help="Open Window displaying preview of the pixelated image")

    multi_processing = args.add_argument_group("Multiprocessing")
    multi_processing.add_argument("--multiprocessing", dest="multiprocessing", action="store_true", help="Enable multi processing for faster CPU execution")
    multi_processing.add_argument("--parallel-processes", dest="parallel_processes", default=4, type=int, help="Number of parallel anonymization processes. One proccess is additional created for displaying preview (if enabled).")

    return args.parse_args()


def main():
    args = parse_args()

    run(**vars(args))


if __name__ == "__main__":
    main()



import os

import numpy as np
import cv2

# ORIGINAL: https://github.com/ORB-HD/deface/blob/master/deface/centerface.py


class CenterFace:
    default_onnx_path = f'{os.path.dirname(__file__)}/centerface.onnx'

    def __init__(self, onnx_path=None, in_shape=None, backend='auto', max_size=640):
        self.in_shape = in_shape
        self.max_size = max_size
        self.onnx_input_name = 'input.1'
        self.onnx_output_names = ['537', '538', '539', '540']

        if onnx_path is None:
            onnx_path = self.default_onnx_path

        if backend == 'auto':
            try:
                import onnx
                import onnxruntime
                backend = 'onnxrt'
            except:
                backend = 'opencv'
        self.backend = backend

        if self.backend == 'opencv':
            self.net = cv2.dnn.readNetFromONNX(onnx_path)
        elif self.backend == 'onnxrt':

            # Silence warnings about unnecessary bn initializers
            onnxruntime.set_default_logger_severity(3)

            static_model = onnx.load(onnx_path)
            dyn_model = self.dynamicize_shapes(static_model)
            self.sess = onnxruntime.InferenceSession(dyn_model.SerializeToString(), providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])

    @staticmethod
    def ensure_rgb(img: np.ndarray) -> np.ndarray:
        """Convert input image to RGB if it is in RGBA or L format"""
        if img.ndim == 2:  # 1-channel grayscale -> RGB
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        elif img.shape[2] == 4:  # 4-channel RGBA -> RGB
            img = cv2.cvtColor(img, cv2.COLOR_RGBA2RGB)
        return img

    @staticmethod
    def dynamicize_shapes(static_model):
        from onnx.tools.update_model_dims import update_inputs_outputs_dims

        input_dims, output_dims = {}, {}
        for node in static_model.graph.input:
            dims = [d.dim_value for d in node.type.tensor_type.shape.dim]
            input_dims[node.name] = dims
        for node in static_model.graph.output:
            dims = [d.dim_value for d in node.type.tensor_type.shape.dim]
            output_dims[node.name] = dims
        input_dims.update({
            'input.1': ['B', 3, 'H', 'W']  # RGB input image
        })
        output_dims.update({
            '537': ['B', 1, 'h', 'w'],  # heatmap
            '538': ['B', 2, 'h', 'w'],  # scale
            '539': ['B', 2, 'h', 'w'],  # offset
            '540': ['B', 10, 'h', 'w']  # landmarks
        })
        dyn_model = update_inputs_outputs_dims(static_model, input_dims, output_dims)
        return dyn_model

    def __call__(self, img, threshold=0.5):
        img = CenterFace.ensure_rgb(img)
        orig_shape = img.shape[:2]
        in_shape = self.in_shape
        if in_shape is None:
            in_shape = orig_shape[::-1]
        w_new, h_new, scale_w, scale_h = self.transform(in_shape, orig_shape)

        blob = cv2.dnn.blobFromImage(
            img, scalefactor=1.0, size=(w_new, h_new),
            mean=(0, 0, 0), swapRB=False, crop=False
        )
        if self.backend == 'opencv':
            self.net.setInput(blob)
            heatmap, scale, offset, lms = self.net.forward(self.onnx_output_names)
        elif self.backend == 'onnxrt':
            heatmap, scale, offset, lms = self.sess.run(self.onnx_output_names, {self.onnx_input_name: blob})
        else:
            raise RuntimeError(f'Unknown backend {self.backend}')
        dets, lms = self.decode(heatmap, scale, offset, lms, (h_new, w_new), threshold=threshold)
        if len(dets) > 0:
            dets[:, 0:4:2], dets[:, 1:4:2] = dets[:, 0:4:2] / scale_w, dets[:, 1:4:2] / scale_h
            lms[:, 0:10:2], lms[:, 1:10:2] = lms[:, 0:10:2] / scale_w, lms[:, 1:10:2] / scale_h
        else:
            dets = np.empty(shape=[0, 5], dtype=np.float32)
            lms = np.empty(shape=[0, 10], dtype=np.float32)

        return dets, lms

    def transform(self, in_shape, orig_shape):
        h_orig, w_orig = orig_shape
        w_new, h_new = in_shape
        if w_new > self.max_size or h_new > self.max_size:
            if w_new > h_new:
                h_new = h_new * (self.max_size / w_new)
                w_new = self.max_size
            else:
                w_new = w_new * (self.max_size / h_new)
                h_new = self.max_size
        # Make spatial dims divisible by 32
        w_new, h_new = int(np.ceil(w_new / 32) * 32), int(np.ceil(h_new / 32) * 32)
        scale_w, scale_h = w_new / w_orig, h_new / h_orig
        return w_new, h_new, scale_w, scale_h

    def decode(self, heatmap, scale, offset, landmark, size, threshold=0.1):
        heatmap = np.squeeze(heatmap)
        scale0, scale1 = scale[0, 0, :, :], scale[0, 1, :, :]
        offset0, offset1 = offset[0, 0, :, :], offset[0, 1, :, :]
        c0, c1 = np.where(heatmap > threshold)
        boxes, lms = [], []
        if len(c0) > 0:
            for i in range(len(c0)):
                s0, s1 = np.exp(scale0[c0[i], c1[i]]) * 4, np.exp(scale1[c0[i], c1[i]]) * 4
                o0, o1 = offset0[c0[i], c1[i]], offset1[c0[i], c1[i]]
                s = heatmap[c0[i], c1[i]]
                x1, y1 = max(0, (c1[i] + o1 + 0.5) * 4 - s1 / 2), max(0, (c0[i] + o0 + 0.5) * 4 - s0 / 2)
                x1, y1 = min(x1, size[1]), min(y1, size[0])
                boxes.append([x1, y1, min(x1 + s1, size[1]), min(y1 + s0, size[0]), s])
                lm = []
                for j in range(5):
                    lm.append(landmark[0, j * 2 + 1, c0[i], c1[i]] * s1 + x1)
                    lm.append(landmark[0, j * 2, c0[i], c1[i]] * s0 + y1)
                lms.append(lm)
            boxes = np.asarray(boxes, dtype=np.float32)
            lms = np.asarray(lms, dtype=np.float32)
            keep = self.nms(boxes[:, :4], boxes[:, 4], 0.3)
            boxes = boxes[keep, :]
            lms = lms[keep, :]
        return boxes, lms

    @staticmethod
    def nms(boxes, scores, nms_thresh):
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]
        areas = (x2 - x1) * (y2 - y1)
        order = np.argsort(scores)[::-1]
        num_detections = boxes.shape[0]
        suppressed = np.zeros((num_detections,), dtype=np.uint8)
        for _i in range(num_detections):
            i = order[_i]
            if suppressed[i]:
                continue
            ix1 = x1[i]
            iy1 = y1[i]
            ix2 = x2[i]
            iy2 = y2[i]
            iarea = areas[i]

            for _j in range(_i + 1, num_detections):
                j = order[_j]
                if suppressed[j]:
                    continue

                xx1 = max(ix1, x1[j])
                yy1 = max(iy1, y1[j])
                xx2 = min(ix2, x2[j])
                yy2 = min(iy2, y2[j])
                w = max(0, xx2 - xx1)
                h = max(0, yy2 - yy1)

                inter = w * h
                ovr = inter / (iarea + areas[j] - inter)
                if ovr >= nms_thresh:
                    suppressed[j] = True
        keep = np.nonzero(suppressed == 0)[0]
        return keep


from setuptools import setup, find_packages
import versioneer

with open('README.md', 'r', encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='PixelateMe',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Marius Butz',
    description='A Python CLI-Tool and package to pixelate or blur faces in images and videos.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mbpictures/pixelateme',
    packages=find_packages(include=["pixelateme"]),
    entry_points={
        'console_scripts': [
            'pixelateme = pixelateme.cli:main',
        ],
        'gui_scripts': [
            'pixelatemegui = pixelateme.gui:main'
        ]
    },
    package_data={'pixelateme': ['centerface.onnx']},
    include_package_data=True,
    install_requires=[
        'deepface',
        'tqdm',
        'opencv-python',
        'onnx',
        'onnxruntime',
        'gooey'
    ],
    extras_require={
        'gpu':  ['onnxruntime-gpu'],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
